"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.prepareNewOrderV1 = exports.Tonic = void 0;
const utils_1 = require("@tonic-foundation/utils");
const storage_1 = require("@tonic-foundation/storage");
const token_1 = require("@tonic-foundation/token");
const bn_js_1 = __importDefault(require("bn.js"));
const batch_1 = require("./batch");
const contract_1 = require("./contract");
const market_1 = require("./market");
const token_2 = require("./token");
const types_1 = require("./types");
const util_1 = require("./util");
const ONE_TGAS = new bn_js_1.default(Math.pow(10, 12));
const MAX_GAS = tgasAmount(300);
const NEAR_DECIMALS = 24;
function tgasAmount(tgas) {
    return new bn_js_1.default(tgas).mul(ONE_TGAS);
}
function nearAmount(amount) {
    return (0, utils_1.decimalToBn)(amount, NEAR_DECIMALS);
}
class Tonic {
    constructor(account, contractId) {
        this.account = account;
        this.contractId = contractId;
        this._contract = (0, contract_1.getContract)(account, contractId);
    }
    get contract() {
        return this._contract;
    }
    /**
     * This function calls a change method, returning both the return value of the
     * call and the execution outcome. Provide the `contractId` property to call a
     * different contract, defaults to the Tonic DEX.
     *
     * Many use cases need transaction/receipt IDs after calling a change
     * method, which makes the NEAR SDK's Contract class unsuitable.
     * https://github.com/near/near-api-js/blob/c84aef6cc18fe5ac1063411c95283ba3130fa656/src/contract.ts#L145
     */
    functionCallWithOutcome(params) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield (0, utils_1.functionCallWithOutcome)(this.account, Object.assign({ contractId: this.contractId }, params));
        });
    }
    /**
     * Execute batch actions. See the example in the "examples" directory for more
     * information.
     */
    executeBatch(batch) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.functionCallWithOutcome({
                methodName: 'execute',
                args: {
                    actions: batch.prepare(),
                },
                gas: tgasAmount(300),
                attachedDeposit: new bn_js_1.default(1),
            });
        });
    }
    /**
     * Deposit NEAR for storage staking. We recommend depositing 0.1 NEAR at a
     * time to give users some room to place orders.
     *
     * Pass `amount` as a decimal amount of NEAR to deposit, eg, 0.1.
     *
     * Passing `registrationOnly` causes deposit in excess of the minimum required
     * balance to be refunded. The minimum balance is computed at runtime, so it's
     * recommended to pass at least 0.01 NEAR to avoid issues.
     *
     * The account to deposit for can be specified with `accountId`. Omit it to
     * deposit for the caller's account.
     */
    storageDeposit(params) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield (0, storage_1.storageDeposit)(this.account, this.contractId, params);
        });
    }
    /**
     * Create a market
     */
    createMarket(params) {
        return __awaiter(this, void 0, void 0, function* () {
            const base_token = (0, token_2.toInternalTokenId)(params.baseTokenId);
            const quote_token = (0, token_2.toInternalTokenId)(params.quoteTokenId);
            const args = {
                base_token,
                quote_token,
                base_token_lot_size: params.baseTokenLotSize,
                quote_token_lot_size: params.quoteTokenLotSize,
                taker_fee_base_rate: params.takerFeeBaseRate,
                maker_rebate_base_rate: params.makerRebateBaseRate,
            };
            return yield this.functionCallWithOutcome({
                methodName: 'create_market',
                args: { args },
                gas: MAX_GAS,
                attachedDeposit: nearAmount(0.1),
            });
        });
    }
    /**
     * Place an order. This is a relatively low-level method that works with
     * lot-aware values. Client applications may use Market.placeOrder instead for
     * a higher-level interface.
     */
    placeOrder(market_id, order) {
        return __awaiter(this, void 0, void 0, function* () {
            const { response: r, executionOutcome } = yield this.functionCallWithOutcome({
                methodName: 'new_order',
                args: {
                    market_id,
                    order: prepareNewOrderV1(order),
                },
                gas: MAX_GAS,
            });
            return {
                response: {
                    id: r.id,
                    outcome: r.outcome,
                    open_quantity: new bn_js_1.default(r.open_quantity),
                    base_fill_quantity: new bn_js_1.default(r.base_fill_quantity),
                    quote_fill_quantity: new bn_js_1.default(r.quote_fill_quantity),
                    base_cancelled_quantity: new bn_js_1.default(r.base_cancelled_quantity),
                },
                executionOutcome,
            };
        });
    }
    /**
     * Cancel an order.
     */
    cancelOrder(market_id, order_id) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.functionCallWithOutcome({
                methodName: 'cancel_order',
                args: { market_id, order_id },
                gas: tgasAmount(100),
            });
        });
    }
    /**
     * Cancel multiple orders in a market. Use Tonic.executeBatch
     * to cancel orders in multiple markets.
     */
    cancelOrders(marketId, orderIds) {
        return __awaiter(this, void 0, void 0, function* () {
            const batch = new batch_1.BatchActionV1();
            batch.cancelOrdersV1(marketId, orderIds);
            return yield this.executeBatch(batch);
        });
    }
    /**
     * Cancel all orders in a market.
     */
    cancelAllOrders(market_id) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.functionCallWithOutcome({
                methodName: 'cancel_all_orders',
                args: { market_id },
                gas: MAX_GAS,
            });
        });
    }
    /**
     * Deposit NEP 141 tokens or native NEAR. For native NEAR, the `tokenId` must
     * be 'NEAR'.
     *
     * The amount must be decimal-aware, eg, to deposit 1 NEAR (24 decimals), use
     *
     * ```
     * import { decimalToBn } from '@tonic-foundation/utils';
     * // Deposit 1 NEAR
     * tonic.deposit('NEAR', decimalToBn(1, 24))
     * // Deposit 1 USDC
     * tonic.deposit('usdc.example.testnet', decimalToBn(1, 6))
     * ```
     */
    deposit(tokenId, amount) {
        return __awaiter(this, void 0, void 0, function* () {
            if (tokenId.toUpperCase() === 'NEAR') {
                return yield this.depositNear(amount);
            }
            else {
                return this.depositFt(tokenId, amount);
            }
        });
    }
    /**
     * Deposit NEP-141 token.
     */
    depositFt(tokenId, amount, msg) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield (0, token_1.ftTransferCall)(this.account, tokenId, {
                receiverId: this.contractId,
                amount,
                msg,
            });
        });
    }
    swap(tokenId, amount, params) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.depositFt(tokenId, amount, params);
        });
    }
    /**
     * Deposit native NEAR.
     */
    depositNear(amount) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.functionCallWithOutcome({
                methodName: 'deposit_near',
                args: {},
                attachedDeposit: amount,
                gas: MAX_GAS,
            });
        });
    }
    /**
     * Withdraw NEP 141 tokens or native NEAR. For native NEAR, the `tokenId` must
     * be 'NEAR'.
     *
     * The amount must be decimal-aware, eg,
     *
     * ```
     * import { decimalToBn } from '@tonic-foundation/utils';
     * // Withdraw 1 NEAR
     * tonic.withdraw('NEAR', decimalToBn(1, 24))
     * // Withdraw 1 USDC
     * tonic.withdraw('usdc.example.testnet', decimalToBn(1, 6))
     * ```
     */
    withdraw(tokenId, amount) {
        return __awaiter(this, void 0, void 0, function* () {
            if (tokenId.toUpperCase() === 'NEAR') {
                return yield this.withdrawNear(amount);
            }
            else {
                return this.withdrawFt(tokenId, amount);
            }
        });
    }
    withdrawFt(tokenId, amount) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.functionCallWithOutcome({
                methodName: 'withdraw_ft',
                args: {
                    token: tokenId,
                    amount: amount.toString(),
                },
                gas: MAX_GAS,
                attachedDeposit: new bn_js_1.default(1),
            });
        });
    }
    withdrawNear(amount) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.functionCallWithOutcome({
                methodName: 'withdraw_near',
                args: {
                    amount: amount.toString(),
                },
                gas: MAX_GAS,
                attachedDeposit: new bn_js_1.default(1),
            });
        });
    }
    /**
     * Get the orderbook. Orders are returned with decimals (eg, 1 USDC is
     * returned as 1000000). The format of each order is [price, open quantity].
     *
     * @param market_id
     * @param depth Number of price levels to fetch, default 12
     * @returns
     */
    getOrderbook(market_id, depth = 12) {
        return __awaiter(this, void 0, void 0, function* () {
            const view = yield this._contract.get_orderbook({
                market_id,
                depth,
            });
            const { bids, asks } = (0, types_1.toOrderBook)(view);
            return {
                bids: (0, util_1.groupOrdersByPriceLevel)(bids),
                asks: (0, util_1.groupOrdersByPriceLevel)(asks),
            };
        });
    }
    getOpenOrders(market_id) {
        return __awaiter(this, void 0, void 0, function* () {
            const raw = yield this._contract.get_open_orders({
                market_id,
                account_id: this.account.accountId,
            });
            return raw.map(types_1.toOpenLimitOrder);
        });
    }
    getOrder(market_id, order_id) {
        return __awaiter(this, void 0, void 0, function* () {
            const raw = yield this._contract.get_order({
                market_id,
                order_id,
            });
            if (!raw) {
                return null;
            }
            return (0, types_1.toOpenLimitOrder)(raw);
        });
    }
    getMarket(marketId) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield market_1.Market.load(this, marketId, this.contractId);
        });
    }
    /**
     * List markets in order of creation.
     */
    listMarkets(offset = 0, limit = 100) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.contract.list_markets({ from_index: offset, limit });
        });
    }
    /**
     * Return available exchange balances with decimals as a map of { tokenId =>
     * balance }. "Available balance" is balance not currently locked in an open
     * order. For example, an account with 1 USDC available would see
     *
     * ```
     * await tonic.getBalances() // { usdc.example.testnet: BN<1000000> }
     * ```
     */
    getBalances() {
        return __awaiter(this, void 0, void 0, function* () {
            const ret = {};
            const rawBalances = yield this._contract.get_balances({
                account_id: this.account.accountId,
            });
            rawBalances.forEach(([token_id, amount]) => {
                ret[(0, token_2.extractTokenContractId)(token_id)] = new bn_js_1.default(amount);
            });
            return ret;
        });
    }
}
exports.Tonic = Tonic;
function prepareNewOrderV1(params) {
    var _a, _b;
    return Object.assign(Object.assign({}, params), { max_spend: (_a = params.max_spend) === null || _a === void 0 ? void 0 : _a.toString(), limit_price: (_b = params.limit_price) === null || _b === void 0 ? void 0 : _b.toString(), quantity: params.quantity.toString() });
}
exports.prepareNewOrderV1 = prepareNewOrderV1;
